from matplotlib import pyplot as plt
import matplotlib.animation as animation
import numpy as np
import random
from IPython.display import HTML
import os
import requests
import zipfile

import torch
import torch.nn as nn
from torch.nn import functional as F
from torchvision import datasets
import torchvision.transforms as transforms
import torchvision.utils as vutils
from tqdm.notebook import tqdm
import mlflow


# ########## TUTORIAL TOOLS ################


def download_tutorial_08(target_file='tutorial_08_files.zip'):
    source_url = f'https://is.muni.cz/el/fi/jaro2024/PA228/um/exercises/tutorial_08/{target_file}'
    print(f'Downloading data from {source_url} to {target_file}')
    if not os.path.exists(target_file):
        print(f'\tDownloading ... {source_url}')
        content = requests.get(source_url).content
        with open(target_file, 'wb') as file:
            file.write(content)
    else:
         print(f'\tFile already downloaded')   
    folder_name = target_file.replace('.zip', '')
    if not os.path.exists(folder_name):
        print(f'\tExtracting to ... {folder_name}')
        with zipfile.ZipFile(target_file, 'r') as zip_ref:
            zip_ref.extractall(folder_name)
    else:
        print(f'\tFile already extracted') 


# ####### AE MODELS #######
class Encoder(nn.Module):
    def __init__(self, latent_dims):
        super(Encoder, self).__init__()
        self.linear1 = nn.Linear(784, 512)
        self.linear2 = nn.Linear(512, latent_dims)
    
    def forward(self, x):
        x = torch.flatten(x, start_dim=1)
        x = F.relu(self.linear1(x))
        return self.linear2(x)
    
class Decoder(nn.Module):
    def __init__(self, latent_dims):
        super(Decoder, self).__init__()
        self.linear1 = nn.Linear(latent_dims, 512)
        self.linear2 = nn.Linear(512, 784)
        
    def forward(self, z):
        z = F.relu(self.linear1(z))
        z = torch.sigmoid(self.linear2(z))
        return z.reshape((-1, 1, 28, 28))

class Autoencoder(nn.Module):
    def __init__(self, latent_dim):
        super(Autoencoder, self).__init__()
        self.latent_dim = latent_dim
        self.encoder = Encoder(latent_dim)
        self.decoder = Decoder(latent_dim)
    
    def forward(self, x):
        z = self.encoder(x)
        return self.decoder(z)

# AUTOENCODER CNN
class ConvBlock(nn.Module):
    def __init__(self,
                 in_channels,
                 out_chanels,
                 activate=True,
                 **kwargs):
        super(ConvBlock, self).__init__()
        self.conv = nn.Conv2d(in_channels, out_chanels, **kwargs)
        self.bn = nn.BatchNorm2d(out_chanels)
        self.activate = activate
        
    def forward(self, x):
        x = self.bn(self.conv(x))
        return F.relu(x) if self.activate else x

    
class EncoderCNN(nn.Module):
    def __init__(self, latent_dims):
        super(EncoderCNN, self).__init__()
        
        self.conv2d1 = ConvBlock(1, 64, kernel_size=(3, 3), padding=1)
        self.conv2d2 = ConvBlock(64, 64, kernel_size=(3, 3), padding=1)
        self.pool1   = nn.MaxPool2d(2)
        self.pool2   = nn.MaxPool2d(2)
        self.linear = nn.Linear(7*7*64, latent_dims)

    
    def forward(self, x):
        x = self.pool1(self.conv2d1(x))        
        x = self.pool2(self.conv2d2(x))
        x = torch.flatten(x, start_dim=1)
        return self.linear(x)
    
class DecoderCNN(nn.Module):
    def __init__(self, latent_dims):
        super(DecoderCNN, self).__init__()
        self.linear = nn.Linear(latent_dims, 7*7*64)
        self.conv2d1 = ConvBlock(64, 64, kernel_size=(3, 3), padding=1)
        self.conv2d2 = ConvBlock(64, 1, kernel_size=(3, 3), padding=1, activate=False)

        
    def forward(self, z):
        z = F.relu(self.linear(z))
        z = z.view(-1, 64, 7, 7)
        z = self.conv2d1(nn.functional.interpolate(z, scale_factor=2, mode='bilinear'))
        z = self.conv2d2(nn.functional.interpolate(z, scale_factor=2, mode='bilinear'))
        
        return torch.sigmoid(z)
        
class AutoencoderCNN(nn.Module):
    def __init__(self, latent_dims):
        super(AutoencoderCNN, self).__init__()
        self.encoder = EncoderCNN(latent_dims)
        self.decoder = DecoderCNN(latent_dims)
    
    def forward(self, x):
        z = self.encoder(x)
        return self.decoder(z)

# ####### VAE MODELS #######

# linear version of VAE
class NormalDistribution():
    def __init__(self, dev='cpu'):
        self.N = torch.distributions.Normal(0, 1)
        self.N.loc = self.N.loc.to(dev) # get sampling on the device
        self.N.scale = self.N.scale.to(dev)
        
    def sample(self, shape):
        return self.N.sample(shape)
        

class ProbabilisticEncoder(nn.Module):
    def __init__(self, latent_dims, dev='cpu'):
        super(ProbabilisticEncoder, self).__init__()
        self.linear1 = nn.Linear(784, 512)
        self.linear_mu = nn.Linear(512, latent_dims)
        self.linear_sigma = nn.Linear(512, latent_dims)
        
        self.latent_dims = latent_dims
        self.mu = None
        self.sigma = None
        
        self.norm_dist = NormalDistribution(dev)
    
    def forward(self, x):
        x = torch.flatten(x, start_dim=1)
        x = F.relu(self.linear1(x))
        self.mu =  self.linear_mu(x)
        self.sigma = torch.exp(self.linear_sigma(x))
        
        # sample distribution
        z = self.mu + self.sigma * self.norm_dist.sample(self.mu.shape)
        
        return z
    
class ProbabilisticDecoder(nn.Module):
    def __init__(self, latent_dims, dev='cpu'):
        super(ProbabilisticDecoder, self).__init__()
        self.linear1 = nn.Linear(latent_dims, 512)
        self.linear_mu = nn.Linear(512, 28*28)
        self.linear_sigma = nn.Linear(512, 28*28)
        
        self.norm_dist = NormalDistribution(dev)
        
    def forward(self, z):
        z = F.relu(self.linear1(z))
        mu =  self.linear_mu(z)
        sigma = torch.exp(self.linear_sigma(z))
        
        # sample distribution
        x_hat = torch.sigmoid(mu + sigma * self.norm_dist.sample(mu.shape))
        return x_hat.view(-1, 1, 28, 28)
                      
                          
class VariationalAutoencoder(nn.Module):
    def __init__(self, latent_dim, dev='cpu'):
        super(VariationalAutoencoder, self).__init__()
        self.latent_dim = latent_dim
        self.encoder = ProbabilisticEncoder(latent_dim, dev)
        self.decoder = ProbabilisticDecoder(latent_dim, dev)
    
    def forward(self, x):
        z = self.encoder(x)
        return self.decoder(z)

class ProbabilisticEncoderCNN(nn.Module):
    def __init__(self, latent_dims, dev='cpu'):
        super(ProbabilisticEncoderCNN, self).__init__()
        
        self.conv2d1 = ConvBlock(1, 64, kernel_size=(3, 3), padding=1)
        self.conv2d2 = ConvBlock(64, 64, kernel_size=(3, 3), padding=1)
        self.pool1   = nn.MaxPool2d(2)
        self.pool2   = nn.MaxPool2d(2)
        self.linear_mu = nn.Linear(7*7*64, latent_dims)
        self.linear_sigma = nn.Linear(7*7*64, latent_dims)
        
        self.latent_dims = latent_dims
        self.mu = None
        self.sigma = None
        
        self.norm_dist = NormalDistribution(dev)
    
    def forward(self, x):
        
        x = self.pool1(self.conv2d1(x))        
        x = self.pool2(self.conv2d2(x))
        x = torch.flatten(x, start_dim=1)
        self.mu =  self.linear_mu(x)
        self.sigma = torch.exp(self.linear_sigma(x))
        
        # sample distribution
        z = self.mu + self.sigma * self.norm_dist.sample(self.mu.shape)
        
        return z
    

class ProbabilisticDecoderCNN(nn.Module):
    def __init__(self, latent_dims, dev='cpu'):
        super(ProbabilisticDecoderCNN, self).__init__()
        self.linear = nn.Linear(latent_dims, 7*7*64)
        self.conv2d1 = ConvBlock(64, 64, kernel_size=(3, 3), padding=1)
        self.conv2d2 = ConvBlock(64, 1, kernel_size=(3, 3), padding=1)
        
        self.linear_mu = nn.Linear(28*28, 28*28)
        self.linear_sigma = nn.Linear(28*28, 28*28)
        
        self.latent_dims = 28*28
        self.mu = None
        self.sigma = None
        
        self.norm_dist = NormalDistribution(dev)

        
    def forward(self, z):
        
        z = F.relu(self.linear(z))
        z = z.view(-1, 64, 7, 7)
        z = self.conv2d1(nn.functional.interpolate(z, scale_factor=2, mode='bilinear'))
        z = self.conv2d2(nn.functional.interpolate(z, scale_factor=2, mode='bilinear'))
        z = torch.flatten(z, start_dim=1)
        
        self.mu =  self.linear_mu(z)
        self.sigma = torch.exp(self.linear_sigma(z))
        
        # sample distribution
        x_hat = self.mu + self.sigma * self.norm_dist.sample(self.mu.shape)   
        return torch.sigmoid(x_hat.view(-1, 1, 28, 28))
    
    
class VariationalAutoencoderCNN(nn.Module):
    def __init__(self, latent_dim, dev='cpu'):
        super(VariationalAutoencoderCNN, self).__init__()
        
        self.latent_dim = latent_dim
        self.encoder = ProbabilisticEncoderCNN(latent_dim, dev)
        self.decoder = ProbabilisticDecoderCNN(latent_dim, dev)
    
    def forward(self, x):
        z = self.encoder(x)
        return self.decoder(z)


# ####### AE TRAINING #######
def batch_loss_ae(model, xb, loss_func, opt=None):
    
    x_hat = model(xb)
    loss = loss_func(xb, x_hat)

    # propagate loss
    if opt is not None:
        loss.backward()
        opt.step()
        opt.zero_grad()

    return loss

def epoch_ae(model,
               data,
               loss_func,
               dev='cpu',
               opt=None,
               prefix=''):
    
    # output structure
    loss_dict = {f'{prefix}loss': 0}
    
    # iterate over data
    for xb, _ in data:
        loss = batch_loss_ae(model, xb.to(dev), loss_func, opt=opt)
        loss_dict[f'{prefix}loss'] += loss.item() * len(xb)

    # normalize by the total number of samples
    loss_dict[f'{prefix}loss'] /= len(data.dataset) 

    return loss_dict
   

def train_ae(epochs, model, loss_func, opt, train_dl, valid_dl, dev):
    for epoch in tqdm(range(epochs)):
        
        # training step
        model.train()
        loss = epoch_ae(model, train_dl, loss_func, dev, opt)
   
        # evaluation step
        model.eval()
        val_loss = epoch_ae(model, valid_dl, loss_func, dev, prefix='val_')
        loss.update(val_loss)

        # CALLBACKS
        mlflow.log_metric("loss", loss["loss"], step=epoch)
        mlflow.log_metric("validation_loss", loss["val_loss"], step=epoch)
                
        print(f'epoch {epoch+1}/{epochs}, loss: {loss["loss"] : .05f}, validation loss: {loss["val_loss"]:.05f}')

# ####### VAE MODELS #######

def compute_kl_loss(encoder):
    '''
    encoder contains attributes self.mu and self.sigma with shape (batch_size, latent_dim)
    '''
    #return ((encoder.sigma**2 + encoder.mu**2) / 2 - torch.log(encoder.sigma) - 1/2).sum
    return - 0.5 * torch.sum(1 + torch.log(encoder.sigma) - encoder.mu**2 - encoder.sigma)
    
def train_vae(epochs, model, loss_func, opt, train_dl, valid_dl, dev, beta=0.01):
    for epoch in tqdm(range(epochs)):
        
        # training step
        model.train()
        loss = epoch_vae(model, train_dl, loss_func, dev, opt, beta=beta)
   
        # evaluation step
        model.eval()
        val_loss = epoch_vae(model, valid_dl, loss_func, dev, prefix='val_', beta=beta)
        loss.update(val_loss)

        # CALLBACKS
        mlflow.log_metric("loss", loss["loss"], step=epoch)
        mlflow.log_metric("validation_loss", loss["val_loss"], step=epoch)
                
        print(f'epoch {epoch+1}/{epochs}, loss: {loss["loss"] : .05f}, validation loss: {loss["val_loss"]:.05f}')
        
        
def batch_loss_vae(model, xb, loss_func, opt=None, beta=0.01):
    
    # pass data
    xb_hat = model(xb)
    b_size = len(xb)
    
    # compute loss
    loss_encoder = compute_kl_loss(model.encoder)
    loss_kl =  loss_encoder / b_size / model.latent_dim    # mean over samples, mean over batch
    loss_rec = loss_func(xb, xb_hat)
    loss = loss_rec + loss_kl * beta

    # propagate loss
    if opt is not None:
        loss.backward()
        opt.step()
        opt.zero_grad()

    return loss_rec, loss_kl


def epoch_vae(model, data, loss_func, dev='cpu', opt=None, beta=0.01, prefix=''):
    
    # create output structure
    loss_dict = {f'{prefix}{name}' : 0 for name in ['loss', 'kl', 'reconstruction'] }
    
    for xb, _ in data:
        
        # compute batch loss
        b_size = len(xb)
        loss_rec, loss_kl = batch_loss_vae(model, xb.to(dev), loss_func, opt=opt, beta=beta)
        loss = loss_rec + loss_kl * beta

        # update stats
        loss_dict[f'{prefix}loss'] += loss.item() * b_size
        loss_dict[f'{prefix}kl'] += loss_kl.item() * beta * b_size
        loss_dict[f'{prefix}reconstruction'] += loss_rec.item() * b_size
        
    # normalize by the numer of samples
    n_samples = len(data.dataset) 
    loss_dict = {key: v / n_samples for key, v in loss_dict.items()}

    return loss_dict

# ####### LATENT SPACE FUNCTIONS #######
def plot_in_latentspace(autoencoder, data, num_batches=15, device='cpu', dims=(0, 1)):
    
    X_list = []
    y_list = []
    for i, (x, y) in enumerate(data):
        if i >= num_batches:
            break
        z = autoencoder.encoder(x.to(device))
        z = z.to('cpu').detach().numpy()
        y = y.to('cpu').detach().numpy()
        X_list.append(z)
        y_list.append(y)
        

            
    X = np.concatenate(X_list, axis=0)
    labels = np.stack(y_list, axis=0)
    print(X.shape, labels.shape)
    centroids = {}
    bboxes = {}
    # Print cluster centres
    for l in range(10):
        centroids[l] = np.mean(X[labels.reshape(-1) == l, :], axis=0)
        bboxes[l] = (np.min(X[labels.reshape(-1) == l, :], axis=0), np.max(X[labels.reshape(-1) == l, :], axis=0))
    d1, d2 = dims
    n_dims = X.shape[1]
    
    plt.figure(figsize=(12, 12))
    plt.scatter(X[:, d1], X[:, d2], c=y_list, cmap='tab10')
    plt.title(f'2D Projection of {n_dims}D Latentspace')
    plt.xlabel(f'dimension idx {d1}')
    plt.ylabel(f'dimension idx {d2}')
    plt.colorbar()
    plt.axis('equal')
    plt.show()
    return centroids, bboxes

def explore_nd_latentspace(autoencoder, centroids, label1, n=3, device='cpu'):

    coeffs = np.linspace(0, 1, n)
    imgs = []
    for l in range(10):
        temp = []
        for c in coeffs:
            z = c*centroids[l] + (1-c)*centroids[label1]
            x_hat = autoencoder.decoder(torch.from_numpy(z).to('cuda'))
            x_hat = x_hat.reshape(28, 28).to('cpu').detach().numpy()
            temp.append(x_hat)
        imgs.append(np.hstack(temp))
            
    plt.figure(figsize=(12, 12))
    #plt.imshow(np.hstack(imgs), extent=[*col_lim, *row_lim])
    plt.imshow(np.vstack(imgs))

def sample_latent_space_bbox(autoencoder, bboxes, label, n=3, device='cpu'):

    coeffs = np.linspace(0, 1, n)
    imgs = []
    low, high = bboxes[label]
    for _ in range(n):
        z = np.random.uniform(low, high).astype(np.float32)
        x_hat = autoencoder.decoder(torch.from_numpy(z).to('cuda'))
        x_hat = x_hat.reshape(28, 28).to('cpu').detach().numpy()
        imgs.append(x_hat)

    plt.figure(figsize=(12, 1))
    #plt.imshow(np.hstack(imgs), extent=[*col_lim, *row_lim])
    plt.imshow(np.hstack(imgs))
            
def sample_latent_space(model, mu=0, sigma=1, n_samples=10, dev='cpu'):
    
    latent_dim = model.latent_dim
    
    if type(mu) == int:
        mu = torch.ones(latent_dim) * mu
    if type(sigma) == int:
        sigma = torch.ones(latent_dim) * sigma
    
    N = torch.distributions.Normal(0, 1)
    
    res = []
    
    for _ in range(n_samples):
        
        z = mu + sigma * N.sample([1, latent_dim])
        res.append(model.decoder(z.to(dev)).detach().cpu().numpy())
        
    return np.concatenate(res)

def plot_samples(samples, grid_dims=(10, 10) ):
    
    w = 28
    
    d0, d1 = grid_dims
    img = np.zeros((d0*w, d1*w))
    
    for i in range(d0):
        for j in range(d1):
            
            img[(d0-1-i)*w:(d0-1-i+1)*w, j*w:(j+1)*w] = samples[i*d0 + j][0]
            
    plt.figure(figsize=(12, 12))
    plt.imshow(img)

# ####### 1D GAN #######

class Dis_1D(nn.Module):
    def __init__(self, sample_size):
        super(Dis_1D, self).__init__()           
        self.main = nn.Sequential(   
            #
            nn.Linear(sample_size, 32, bias=True),
            nn.LeakyReLU(0.1, inplace=True),
            #
            nn.Linear(32, 1, bias=True), 
            nn.Sigmoid(),
        )
    def forward(self, input): return self.main(input)

class Gen_1D(nn.Module):
    def __init__(self, sample_size, latent_dim):
        super(Gen_1D, self).__init__()          
        self.main = nn.Sequential(  
            #
            nn.Linear(latent_dim, 32, bias=True),
            nn.LeakyReLU(0.3, inplace=True),
            #
            nn.Linear(32, sample_size, bias=True), 
        )
    def forward(self, input): return self.main(input)

def plot_GAN_1D(real, fake, pred_labels, epoch=None, fig=None):
    if fig is not None:
        fig.clear()
    else:
        plt.figure(figsize=(5,4))
    plt.hist((fake, real), density=True, histtype='barstacked', bins=10,
             color = ['blue', 'orange'])
    plt.scatter(np.arange(-20, 20, 0.5), pred_labels, c='r')
    plt.xlim(-20, 20)
    plt.ylim(0, 1)
    if epoch is not None: 
        plt.title(f'epoch {epoch}')
    plt.legend(['Discriminator labels', 'Fake samples', 'Real samples'])
    if fig is not None:
        fig.canvas.draw()
    else:
        plt.show()

def fit_dg_mix_loss(netD, netG, data, noise, dev, real_label, fake_label, 
                    criterion, optimizerD, optimizerG):
    
    ############################
    # (1) Update D network: maximize log(D(x)) + log(1 - D(G(z)))
    ###########################
    ## Train with all-real batch
    netD.zero_grad()
    # Format batch
    real_cpu = data.to(dev)
    batch_size = real_cpu.size(0)
    label_real = torch.full((batch_size,), real_label, dtype=torch.float, 
                            device=dev)
    # Forward pass real batch through D
    output_real = netD(real_cpu).view(-1)
    
    # Generate fake image batch with G
    fake = netG(noise)
    # Classify all fake batch with D
    output_fake = netD(fake.detach()).view(-1)
    label_fake = torch.full((batch_size,), fake_label, dtype=torch.float, 
                            device=dev)
    
    outputs = torch.cat((output_real, output_fake), 0)
    targets = torch.cat((label_real, label_fake), 0)

    errD = criterion(outputs, targets)
    errD.backward()
       
    D_x = output_real.mean().item()
    D_G_z1 = output_fake.mean().item()
    
    optimizerD.step()

    ############################
    # (2) Update G network: maximize log(D(G(z)))
    ###########################
    netG.zero_grad()
    # Since we just updated D, perform another forward pass of all-fake 
    # batch through D
    output = netD(fake).view(-1)
    #output = output_fake
    label_real = torch.full((batch_size,), real_label, dtype=torch.float,
                            device=dev)
    # Calculate G's loss based on this output
    errG = criterion(output, label_real)
    # Calculate gradients for G
    errG.backward()
    D_G_z2 = output.mean().item()
    # Update G
    optimizerG.step()
    
    return errD, errG, D_x, D_G_z1, D_G_z2

def train_GAN_1D(netD, netG, criterion, optD, optG, dev, prep_real_samples,
                 num_epochs, batch_size, latent_dim, sample_size, fig):  
    
    # reset network weights
    netG.apply(reset_weights_x)
    netD.apply(reset_weights_x)
    # fix random seeds    
    set_seeds(0)
    
    # Establish convention for real and fake labels during training
    real_label = 1.
    fake_label = 0.

    # Lists to keep track of progress
    G_losses = []
    D_losses = []

    # Create batch of latent vectors that we will use to visualize
    # the progression of the generator
    fixed_noise = (torch.rand(batch_size, latent_dim))
    fixed_noise = fixed_noise.to(dev).float()
    
    real_fixed = prep_real_samples(batch_size, sample_size)
    
    print("Training...")
    netD.train()
    netG.train()
    # For each epoch
    iters = 0
    for epoch in tqdm(range(num_epochs), 'epochs'):
        
        data = torch.from_numpy(prep_real_samples(batch_size, 
                                                  sample_size)).float()
        # Generate batch of latent vectors
        noise = torch.rand(batch_size, latent_dim, device=dev)
        
        errD, errG, D_x, D_G_z1, D_G_z2=fit_dg_mix_loss(netD, netG, data,
            noise, dev, real_label, fake_label, criterion, optD, optG)

        if epoch % 250 == 0:
            mlflow.log_metric("Generator loss", errG.item(), step=epoch)
            mlflow.log_metric("Discriminator loss", errD.item(), step=epoch)
        
        if epoch % 100 == 0 or epoch == num_epochs-1:
            with torch.no_grad():      
                # Generate fake image batch with G
                fake = netG(fixed_noise)
                
                fake_flat = np.ravel(fake.detach().cpu().numpy())
                real_flat = np.ravel(real_fixed)
                
                pred_labels = netD(torch.arange(-20, 20, 
                                                0.5).reshape((-1, 1)).to(dev))
                pred_labels = np.ravel(pred_labels.detach().cpu().numpy())
                
                plot_GAN_1D(real_flat, fake_flat, pred_labels, epoch, fig)                
    
        iters += 1
        
    return

# ####### 2D GAN #######

def set_seeds(seed):
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)

# normal distribution initialization
def reset_weights_n(m):
    classname = m.__class__.__name__
    if classname.find('Conv') != -1:
        nn.init.normal_(m.weight.data, 0.0, 0.02)
    elif classname.find('BatchNorm') != -1:
        nn.init.normal_(m.weight.data, 1.0, 0.02)
        nn.init.constant_(m.bias.data, 0)

# xavier initialization
def reset_weights_x(m):
    set_seeds(0)  
    if isinstance(m, nn.Conv2d):
        nn.init.xavier_uniform_(m.weight.data)
        if m.bias is not None:
                nn.init.constant_(m.bias.data, 0)
    if isinstance(m, nn.ConvTranspose2d):
        nn.init.xavier_uniform_(m.weight.data)
        if m.bias is not None:
                nn.init.constant_(m.bias.data, 0)
    elif isinstance(m, nn.BatchNorm2d):
        nn.init.constant_(m.weight.data, 1)
        nn.init.constant_(m.bias.data, 0)
    elif isinstance(m, nn.Linear):
        nn.init.xavier_uniform_(m.weight.data)
        nn.init.constant_(m.bias.data, 0)

def get_data(dataset=None, batch_size=128):
    
    # fix random seeds
    set_seeds(0)
    
    if dataset == 'fmnist':
        # transform images to tensors of normalized range (from [0, 1]
        # to [-1, 1])
        # normalization: output_intensity = (input_intensity - mean) / std
        # in our case mean = 0.5, std = 0.5
        #
        # Note that we need to do this for all three image channels.
        transform = transforms.Compose([
            transforms.ToTensor(),
            transforms.Normalize((0.5), (0.5))])
        
        # training data
        trainset = datasets.FashionMNIST(root='./data', train=True,
                                         download=True, transform=transform)
        # training data loader   
        #                           
        # Note: If training ends with memory error (especially on Windows),
        # make sure to set num_workers=0. Increasing the available workers
        # speeds up the training (especially when doing on-the-fly
        # augmentation) but needs more memory.           
        trainloader = torch.utils.data.DataLoader(trainset,
            batch_size=batch_size, shuffle=True, num_workers=0)
        
        return trainset, trainloader      
        
    else:    
        raise Exception("Unsupported dataset!")

def dataset_preview(dataset, mode=None, show_labels=False):
    
    # fix random seeds
    set_seeds(0)
    
    if show_labels:
        plt.figure(figsize=(7, 7))
    else:
        plt.figure(figsize=(6, 6),facecolor='black')
        
    rand_samples = np.random.permutation(np.arange(0,len(dataset)))
        
    if mode == 'fmnist':
        for i in range(64):
            rand_sample = rand_samples[i]
            plt.subplot(8,8,i+1)    
            plt.imshow(dataset[rand_sample][0].squeeze(), cmap='gray',
                       interpolation='none')    
            if show_labels:
                plt.title("{}".format(
                    dataset.classes[dataset[rand_sample][1]]))
            plt.xticks([])    
            plt.yticks([])
        if show_labels:
            plt.subplots_adjust(wspace=0, hspace=0.5)
        else:
            plt.subplots_adjust(wspace=0, hspace=0)

    else:
        raise Exception("Unsupported dataset!")

    plt.tight_layout()

class Dis_cCNN(nn.Module):
    def __init__(self):
        super(Dis_cCNN, self).__init__()     
        self.ndf = 128 # depth of feature maps propagated through the network  
        self.label_emb = nn.Embedding(10, 10)
        self.fc0 = nn.Linear(10, 28*28)
        self.res = nn.Unflatten(-1, torch.Size([1, 28, 28]))     
        self.main = nn.Sequential(
            #
            nn.Conv2d(2, self.ndf, kernel_size=4, stride=2,
                      padding=3, bias=False),
            nn.LeakyReLU(0.2, inplace=True),
            #
            nn.Conv2d(self.ndf * 1, self.ndf * 2, kernel_size=4, stride=2,
                      padding=1, bias=False),
            nn.BatchNorm2d(self.ndf * 2),
            nn.LeakyReLU(0.2, inplace=True),
            #
            nn.Conv2d(self.ndf * 2, self.ndf * 4, kernel_size=4, stride=2,
                      padding=1, bias=False),
            nn.BatchNorm2d(self.ndf * 4),
            nn.LeakyReLU(0.2, inplace=True),
            #
            nn.Conv2d(self.ndf * 4, 1, kernel_size=4, stride=1,
                      padding=0, bias=False),
            nn.Sigmoid()
        )
    # input shape:[batch_size, 1, 28, 28], labels shape:[batch_size]
    def forward(self, input, labels): 
        c = self.label_emb(labels) # [batch_size] -> [batch_size, 10]
        c = self.fc0(c)  # [batch_size, 10] -> [batch_size, 28*28]
        c = self.res(c) # [batch_size, 28*28] -> [batch_size, 1, 28, 28]
        # [ [batch_size, 1, 28, 28], [batch_size, 1, 28, 28] ]
        # -> [batch_size, 2, 28, 28]
        out = torch.cat([input, c], 1) 
        out =  self.main(out)  
        return out

class Gen_cCNN(nn.Module):
    def __init__(self, latent_dim):
        super(Gen_cCNN, self).__init__()
        self.ngf = 128 # depth of feature maps propagated through the network 
        self.label_emb = nn.Embedding(10, 10)
        self.main = nn.Sequential(
            #
            nn.ConvTranspose2d(latent_dim+10, self.ngf * 4, kernel_size=4,
                               stride=1, padding=0, bias=False),
            nn.BatchNorm2d(self.ngf * 4),
            nn.ReLU(True),
            #
            nn.ConvTranspose2d(self.ngf * 4, self.ngf * 2, kernel_size=4,
                               stride=2, padding=1, bias=False),
            nn.BatchNorm2d(self.ngf * 2),
            nn.ReLU(True),
            #
            nn.ConvTranspose2d(self.ngf * 2, self.ngf * 1, kernel_size=4,
                               stride=2, padding=1, bias=False),
            nn.BatchNorm2d(self.ngf * 1),
            nn.ReLU(True),
            #
            nn.ConvTranspose2d(self.ngf, 1, kernel_size=4,
                               stride=2, padding=3, bias=False),
            nn.Tanh()            
        )
    # input shape:[batch_size, latent_dim, 1, 1], labels shape:[batch_size]
    def forward(self, input, labels):
        # [batch_size] -> [batch_size, 10] -> [batch_size, 10, 1, 1]
        c = self.label_emb(labels).unsqueeze(2).unsqueeze(2) 
        # [ [batch_size, latent_dim, 1, 1], [batch_size, 10, 1, 1] ]
        # -> [batch_size, latent_dim+10, 1, 1]
        out = torch.cat([input, c], 1)
        out = self.main(out)
        return out

def show_interp(tensor_images, n_points):
    plt.figure(figsize=(6,1))
    plt.axis("off")
    plt.title("Generator output: point A → point B")
    plt.tight_layout()
    plt.imshow(np.transpose(vutils.make_grid(tensor_images, nrow = n_points, 
        padding=2, normalize=True).cpu(),(1,2,0)))

def batch2grid(fake):
    return vutils.make_grid(fake, padding=2, normalize=True)

def imgshow(img, cmap='gray', title='', fig_size=(2,2), colorbar=False,
            interpolation='none'):
    fig, ax = plt.subplots(figsize=fig_size)
    pcm = ax.imshow(img, cmap=cmap, interpolation=interpolation)
    ax.set_frame_on(False)
    ax.tick_params(top=True, labeltop=True, bottom=False, labelbottom=False)
    plt.title(title)
    plt.tight_layout()
    if colorbar:
        fig.colorbar(pcm, orientation='vertical')
    plt.show()

def show_img_epochs(img_list):      
    fig = plt.figure(figsize=(6,6))
    plt.axis("off")
    plt.tight_layout()
    fimgs = [[plt.imshow(np.transpose(img_list[i], (1,2,0)), animated=True)] for i in range(len(img_list))]
    plt.close('all')
    ani = animation.ArtistAnimation(fig, fimgs, interval=1000, repeat_delay=1000, blit=True)
    return HTML(ani.to_jshtml())
    
def fit_cdg_mix_loss(netD, netG, data, noise, labels_conditions, dev, 
                     real_label, fake_label, criterion, optimizerD,
                     optimizerG):
    
    ############################
    # (1) Update D network: maximize log(D(x)) + log(1 - D(G(z)))
    ###########################
    ## Train with all-real batch
    netD.zero_grad()
    # Format batch
    labels_conditions = labels_conditions.to(dev)
    real_cpu = data.to(dev)
    batch_size = real_cpu.size(0)
    label_real = torch.full((batch_size,), real_label, dtype=torch.float,
                            device=dev)
    # Forward pass real batch through D
    output_real = netD(real_cpu, labels_conditions).view(-1)
    
    # Generate fake image batch with G
    fake_labels_conditions = torch.tensor(
        np.random.randint(0, 10, batch_size)).to(dev)
    fake = netG(noise, fake_labels_conditions)
    # Classify all fake batch with D
    output_fake = netD(fake.detach(), fake_labels_conditions).view(-1)
    label_fake = torch.full((batch_size,), fake_label, dtype=torch.float,
                            device=dev)
    
    outputs = torch.cat((output_real, output_fake), 0)
    targets = torch.cat((label_real, label_fake), 0)

    errD = criterion(outputs, targets)
    errD.backward()
       
    D_x = output_real.mean().item()
    D_G_z1 = output_fake.mean().item()
    
    optimizerD.step()

    ############################
    # (2) Update G network: maximize log(D(G(z)))
    ###########################
    netG.zero_grad()
    # Since we just updated D, perform another forward pass of all-fake batch
    # through D
    output = netD(fake, fake_labels_conditions).view(-1)
    label_real = torch.full((batch_size,), real_label, dtype=torch.float,
                            device=dev)
    # Calculate G's loss based on this output
    errG = criterion(output, label_real)
    # Calculate gradients for G
    errG.backward()
    D_G_z2 = output.mean().item()
    # Update G
    optimizerG.step()
    
    return errD, errG, D_x, D_G_z1, D_G_z2

def train_GAN_cCNN(netD, netG, criterion, optD, optG, dev,
                 latent_dim, num_epochs, trainloader):  
    
    # reset network weights
    netG.apply(reset_weights_n)
    netD.apply(reset_weights_n)
    # fix random seeds    
    set_seeds(0)
 
    # Establish convention for real and fake labels during training
    real_label = 1.
    fake_label = 0.

    # Lists to keep track of progress
    img_list = []
    G_losses = []
    D_losses = []
 
    # Create batch of latent vectors that we will use to visualize
    # the progression of the generator
    fixed_noise = torch.randn(80, latent_dim, 1, 1, device=dev)
    fixed_labels = torch.LongTensor(
        [i for i in range(10) for _ in range(8)]).to(dev)
    
    print("Training...")
    netD.train()
    netG.train()
    # For each epoch
    iters = 0
    for epoch in tqdm(range(num_epochs), desc='  epochs'):
        for i, (data, labels_conditions) in tqdm(enumerate(trainloader),
            total=len(trainloader), leave=False, desc='iterations'):
            
            batch_size = data.shape[0]
        
            # Generate batch of latent vectors
            noise = torch.randn(batch_size, latent_dim, 1, 1, device=dev) 

            errD, errG, D_x, D_G_z1, D_G_z2 = fit_cdg_mix_loss(netD, netG,
               data, noise, labels_conditions, dev, real_label, fake_label,
               criterion, optD, optG)

            if iters % 50 == 0:
                mlflow.log_metric("Generator loss", errG.item(), step=epoch)
                mlflow.log_metric("Discriminator loss", errD.item(), step=epoch)
                    
            # Check how the generator is doing by saving G's output
            # on fixed_noise
            if (iters % 100 == 0) or ((epoch == num_epochs-1) and 
                                     (i == len(trainloader)-1)):
                with torch.no_grad():
                    netG.eval()
                    fake = netG(fixed_noise, fixed_labels).detach().cpu()
                    netG.train()
                img_list.append(batch2grid(fake))

            iters += 1
        
    return img_list

def get_samples_normal(batch_size, sample_size):

    # parameters of the normal distribution that we want to generate
    real_mean = 8
    real_std = 1
    
    # output size
    sz = [batch_size, sample_size]
    
    # get random samples
    samples = np.random.normal(real_mean, real_std, sz)
        
    # check the output shape
    assert samples.shape == (batch_size, sample_size)
    
    return samples